#include <cstdio>
#include <cstdlib>
#include "smem.hpp"
#include "em.hpp"
#include "kmeans.hpp"
#include "math.h"
#include "timer.hpp"


//implementation of "SMEM algorithm for mixture models, Naonori Ueda, Ryohei Nakano"
//algorithm aims at solving the local maxima problem of EM algorithm.
//based on the EM implementation from Ismo
//author: Qinpei Zhao,  start: 09.09.2008
//first version:   16.11.2008      basic function finished

const float pi = 3.1415;
static std::vector<std::pair<double,Matrix> > DetInvs;
static std::vector<std::pair<double,Matrix> > DetInvsNew;	


//converge with threshold
double SplitMergeEM(MixtureModel& MM, const Data& D,const Matrix& MinimumVariances, 
					int SMEMIterations, double EMthreshold, int cn, bool Verbose)
{	
	FILE* f = fopen("SMEM.txt", "a+");
	double Qvalue_current = 0;
	double Q_convergence = 0;
	double LKcurrent = 0;
	double LKoptimal = 0;

	//candidate for merge and split
	int Cmax = MM.Count()*(MM.Count()-1)*(MM.Count()-2)/2;

    bool Diagonal = true;

    for (int k = 0; Diagonal && k < MM.Count(); ++k)
        if (MM[k].MatrixType() != MM_COV_TYPE_DIAG) Diagonal = false;

	//perform usual EM, got new model and then Q function value
	LKcurrent = ExpectationMaximization(MM, D, MinimumVariances, EMthreshold, Verbose);
	LKoptimal = LKcurrent;
	Qvalue_current = QFunction(MM, D);
	//fprintf(stdout, "initial Likelihood = %f\n", LKcurrent);
	//MM.Write(f, MM_FF_PLAIN);
	Q_convergence = Qvalue_current;
	MixtureModel BestMM = MM;
	for(int i = 0; i < SMEMIterations; ++i)
	{		
		//fprintf(stdout, "iter = %d\n", i);
		//calculate split and merge criteria, get candidates
		//matrix after merge criterion
		Matrix MergeResult(BestMM.Count(), BestMM.Count());
		Merge(BestMM, D, MergeResult);	
		//matrix after split criterion
		Matrix SplitResult(1, BestMM.Count());
		Split(BestMM, D, SplitResult);
			
		Matrix Candidates(Cmax, 3);
		SortingCandidate(MergeResult, SplitResult, Candidates);
//		fprintf(stdout, "sm iteration:%d\n", i);
		// for each candidate
		for(int c = 0; c < cn; ++c)
		{
			MixtureModel Model = BestMM;
			
			int merge1 = Candidates(c,0);
			int merge2 = Candidates(c,1);
			int split =  Candidates(c,2);
			
			PartialEM(BestMM, D, Model, merge1, merge2, split, Diagonal);	
			//Model.Write(f, MM_FF_PLAIN);
			LKcurrent = ExpectationMaximization(Model, D, MinimumVariances, EMthreshold, Verbose);		
			//Model.Write(f, MM_FF_PLAIN);
			Qvalue_current =  QFunction(Model, D);

			if(LKcurrent > LKoptimal  )  //|| (Qvalue_current >= Q_convergence)
			{
				//Model.Write(f, MM_FF_PLAIN);
				
				//fprintf(stdout, "i = %d, j = %d, k = %d\n", merge1, merge2, split);
				//fprintf(stdout, "time: %s; likelihood: %f\n", timer::StringS(timer::Now()), LKcurrent );
				//fprintf(stdout, "It was improved on candidate %d\n", c);
				LKoptimal = LKcurrent;		
				Q_convergence = Qvalue_current;

//				fprintf(stdout, "Loglikelihood = %f; Qvalue = %f\n", LKoptimal, Q_convergence);
				BestMM = Model;
//				fprintf(f, "\n----models after %d success iteration\n\n", i+1);
				
				break;
			}
		}

//		fprintf(stdout, "iteration = %d; Qvalue = %f\n", i, Q_convergence);
		if( LKoptimal > LKcurrent ) // (Q_convergence > Qvalue_current)
			break;
		
	}
	MM = BestMM;
	return LKoptimal; // return Q_convergence ;
}


//function on SMEM algorithm, converge with iterations
double SplitMergeEM(MixtureModel& MM, const Data& D, const Matrix& MinimumVariances,
					int SMEMIterations, int EMIterations, int cn, bool Verbose)
{
	double Qvalue_current = 0;
	double Q_convergence = 0;
	double LKcurrent = 0;
	double LKoptimal = 0;

	//candidate for merge and split
	int Cmax = MM.Count()*(MM.Count()-1)*(MM.Count()-2)/2;

    bool Diagonal = true;
    for (int k = 0; Diagonal && k < MM.Count(); ++k)
        if (MM[k].MatrixType() != MM_COV_TYPE_DIAG) Diagonal = false;

	//perform usual EM, got new model and then Q function value
	LKcurrent = ExpectationMaximization(MM, D, MinimumVariances, EMIterations, Verbose);
	LKoptimal = LKcurrent;
	Qvalue_current = QFunction(MM, D);
	//fprintf(stdout, "original Likelihood = %f\n", LKcurrent);
	Q_convergence = Qvalue_current;
	MixtureModel BestMM = MM;
	for(int i = 0; i < SMEMIterations; ++i)
	{		
		//fprintf(stdout, "iter = %d\n", i);
		//calculate split and merge criteria, get candidates
		//matrix after merge criterion
		Matrix MergeResult(BestMM.Count(), BestMM.Count());
		Merge(BestMM, D, MergeResult);	
		//matrix after split criterion
		Matrix SplitResult(1, BestMM.Count());
		Split(BestMM, D, SplitResult);
			
		Matrix Candidates(Cmax, 3);
		SortingCandidate(MergeResult, SplitResult, Candidates);

		// for each candidate
		for(int c = 0; c < cn; ++c)
		{
			MixtureModel Model = BestMM;
			int merge1 = Candidates(c,0);
			int merge2 = Candidates(c,1);
			int split =  Candidates(c,2);
			PartialEM(BestMM, D, Model, merge1, merge2, split, Diagonal);
			LKcurrent = ExpectationMaximization(Model, D, MinimumVariances, EMIterations);

			Qvalue_current =  QFunction(Model, D);

			if(LKcurrent > LKoptimal  )  //|| (Qvalue_current >= Q_convergence)
			{
				//fprintf(stdout, "time: %s; likelihood: %f\n", timer::StringS(timer::Now()), LKcurrent );
//				fprintf(stdout, "It was improved on candidate %d\n", c);		
				LKoptimal = LKcurrent;
//				fprintf(stdout, "Loglikelihood = %f\n", LKoptimal);
				Q_convergence = Qvalue_current;
				BestMM = Model;
				break;
			}
		}
//		fprintf(stdout, "iteration = %d; Qvalue = %f\n", i, Q_convergence);
		if(LKoptimal > LKcurrent) // (Q_convergence > Qvalue_current )
			break;
		
	}
	MM = BestMM;
	return LKoptimal; // return Q_convergence ;
}

//define the Qfunction used to estimate the goodness of models
double QFunction(const MixtureModel& MM, const Data& D)
{
	double Qvalue = 0;
	double value = 0;

	SetupDeterminantAndInverseMatrices(DetInvs, MM);
	Matrix PosteriorRow(1, MM.Count());
	// weight*probability density(n,m)

	Matrix Single(1, MM.Count());

	for (size_t n = 0; n < D.Count(); ++n)
	{
		ConstRowIterator VectorBegin = D.ConstVectorBegin(n);
		double Denom = 0;
		for (int m = 0; m < MM.Count(); ++m)
		{
			double Mah2;
			if (MM[m].MatrixType() == MM_COV_TYPE_DIAG)
				Mah2 = MahalanobisDiag2(VectorBegin,
						MM[m].Mean().ConstRowBegin(0), DetInvs[m].second);
			else  Mah2 = MahalanobisChol2(VectorBegin,
						MM[m].Mean().ConstRowBegin(0), DetInvs[m].second);
			PosteriorRow[m] = MM[m].Weight() * ProbabilityDensity(Mah2, DetInvs[m].first);
			if(PosteriorRow[m] == 0)
				PosteriorRow[m] = DBL_MIN;

			Single[m] = log(PosteriorRow[m]);		
			Denom += PosteriorRow[m];
	
		}// end for m  MM.Count
		if (Denom < DBL_MIN)
			PosteriorRow.Fill(1.0 / MM.Count());
		else
			PosteriorRow *= 1.0 / Denom; 

		value = Dot(PosteriorRow, Single);

		Qvalue += value;

	}//end for n D.Count

	return Qvalue/D.Count();
}

// define the merge criterion
void Merge(const MixtureModel& MM, const Data& D, Matrix& JMerge)
{
	JMerge.Fill(0.0);
	Matrix PosteriorRow(1, MM.Count());
	Matrix Posterior(D.Count(), MM.Count());
//	JMerge(MM.Count(), MM.Count());
//	JMerge.Fill(0.0);

	SetupDeterminantAndInverseMatrices(DetInvs, MM);
	// posterior probabilities for the all the models. matrix n*m
	for (size_t n = 0; n < D.Count(); ++n)
	{
		ConstRowIterator VectorBegin = D.ConstVectorBegin(n);
		double Denom = 0;
		for (int m = 0; m < MM.Count(); ++m)
		{
			double Mah2;
			if (MM[m].MatrixType() == MM_COV_TYPE_DIAG)
				Mah2 = MahalanobisDiag2(VectorBegin,
						MM[m].Mean().ConstRowBegin(0), DetInvs[m].second);
			else Mah2 = MahalanobisChol2(VectorBegin,
						MM[m].Mean().ConstRowBegin(0), DetInvs[m].second);
			PosteriorRow[m] = MM[m].Weight() * ProbabilityDensity(Mah2, DetInvs[m].first);
			Denom += PosteriorRow[m];
			
		}// end for m  MM.Count  ?? DBL_MIN
		if (Denom < DBL_MIN)
			PosteriorRow.Fill(1.0 / MM.Count());
		else PosteriorRow *= 1.0 / Denom;

		for (int t = 0; t < MM.Count(); ++t)
			Posterior(n,t) = PosteriorRow[t];

	}//end for n D.Count

	//pick up ith and jth model from Matrix N*M (M-components)
	//modified criterion with normalization
	//calculate ||P()||
	for( int i = 0; i < Posterior.Cols(); ++i)
	{
		for( int j = i+1; j < Posterior.Cols(); ++j)
		{
			double sum_ij = 0;
			double norm_pi = 0;
			double norm_pj = 0;
			for (size_t n = 0; n < D.Count(); ++n)
			{
				sum_ij += Posterior(n, i)* Posterior(n, j); 
				norm_pi += Posterior(n, i)* Posterior(n, i);
				norm_pj += Posterior(n, j)* Posterior(n, j);
			}
			norm_pi = sqrt(norm_pi);
			norm_pj = sqrt(norm_pj);
			//calculate the merge criterion value;
			//Transpose(Pi)*Pj;  
			JMerge(i,j) = sum_ij / (norm_pi*norm_pj);
		}
	}//end for calculate JMerge

}

//define the split criterion
void Split(const MixtureModel& MM, const Data& D, Matrix& JSplit)
{	
	JSplit.Fill(0.0);
	SetupDeterminantAndInverseMatrices(DetInvs, MM);
	// local data density around each (m) model
	for (int m = 0; m < MM.Count(); ++m)
	{
		double KLDiverge = 0;
		double Norm = 0;
		Matrix EstimatedProb(D.Count(), 1);
		Matrix ActualProb(D.Count(), 1);
		ActualProb.Fill(0.0);
		Matrix Dist(D.Count(), 1);
		Matrix Index_Sorted(D.Count(), 1);
		for (size_t n = 0; n < D.Count(); ++n)
		{
			ConstRowIterator VectorBegin = D.ConstVectorBegin(n);
			double Mah2;
			if (MM[m].MatrixType() == MM_COV_TYPE_DIAG)
				Mah2 = MahalanobisDiag2(VectorBegin,
						MM[m].Mean().ConstRowBegin(0), DetInvs[m].second);
			else Mah2 = MahalanobisChol2(VectorBegin,
						MM[m].Mean().ConstRowBegin(0), DetInvs[m].second);
			//Empirical distribution density for kth model
			//calculate distance between each data point and the current model		
			Dist[n] = 1/Mah2;
			EstimatedProb[n] = ProbabilityDensity(Mah2, DetInvs[m].first);
			// in case Mah2 is too big to make the prob. = 0; 
			// case: Mah2 = 2430.777456;Cov = 2201655001.606664;
			// I need to avoid EstimatedProb = 0, reason refer to the split criteria!
			if(EstimatedProb[n] == 0)
				EstimatedProb[n] = DBL_MIN;
			Norm += EstimatedProb[n];
		}//end for n D.Count
		// normalize the prob. cause it has different range with ActualProb.
		if (Norm < DBL_MIN)
			EstimatedProb.Fill(1.0 / MM.Count());
		else EstimatedProb *= 1.0 / Norm;	

		//sort 1/Dist[n] in increasing order, get the index after sorting
		//distance smaller means nearer to the model, give more probability.
		//experical distribution
		Matrix MinIndex(D.Count(),1);		
		for(size_t t = 0; t < D.Count(); ++t)
		{	
			size_t CurrIndex = 0;
			double MinDist = DBL_MAX;
			for(size_t n = 0; n < D.Count(); ++n)
			{
				if(Dist[n] != DBL_MAX && Dist[n] < MinDist)
				{
					MinDist = Dist[n];
					CurrIndex = n;
				}	
			}
			MinIndex[t] = CurrIndex;
			Dist[CurrIndex] = DBL_MAX;
		}// finding the sorting index

		for(size_t s = 0; s < D.Count(); ++s)
		{
			//assign probability of each point with the sorting index/D.Count()
			//ground truth density by empirical distribution
			ActualProb[MinIndex[s]] = 2*(s+1)/double(D.Count()*(D.Count()+1));
			if(ActualProb[MinIndex[s]] == 0)
				fprintf(stdout, "there are zeros!!!!\n");
			double temp = ActualProb[MinIndex[s]] * log(ActualProb[MinIndex[s]]/EstimatedProb[MinIndex[s]]); 
			//when EstimatedProb is very very small, then temp is inf. In this case, we set the
			// split value of the current model is highest, which is DBL_MAX. definitely first to split.
			if(temp > DBL_MAX)
			{
				temp = DBL_MAX;
				KLDiverge = temp;
				continue;
			}
			KLDiverge += temp;
			
		}
		JSplit[m] = KLDiverge;
	}// end for m MM.Count
}


// Kernel Density Estimation; this function is used to estimate the probability density of
// of the total population in terms of some samples of the data.
// We try to make it in different kernels: Gaussian is default.
// This function will be used in split function to calculate the local data denstiy. 
// return the probability value of current data sample
/*
double KernelDensity(ConstRowIterator VectorBegin,  const Data &D)
{
	double h = 0.5;
	size_t d = D.Dimensionality();
	double coef = pow(2*pi, d/2) * D.Count() * h;

	ConstRowIterator VectCurr = VectorBegin;
	ConstRowIterator VectorEnd = VectorBegin;
	double value = 0;
	++VectorEnd;
	while(VectorBegin != VectorEnd)
	{
		double tempvalue = 0;
		tempvalue = (*VectCurr - *VectorBegin)*(*VectCurr - *VectorBegin);
		value += exp(-0.5/(h*h)*tempvalue);
		++VectorBegin;
	}
	return value;

}*/


// sort the merge and split criterion value with decreasing order. 
// large value of both indicate good candicate to merge and split
void SortingCandidate(Matrix& JMerge, Matrix& JSplit, Matrix& Candidates)
{
	Candidates.Fill(0.0);
	int M = JMerge.Cols();
	Matrix Split_Backup;
	Matrix Merge_Backup = JMerge;
	int Cmax = M*(M-1)*(M-2)/2;
//	Matrix Candidate(Cmax,3); 
	//find i,j for merging
	//store index after sorting
	Matrix MaxIndex_Merge((M*M-M)/2, 2);
	for(int m = 0; m < (M*M-M)/2; ++m)
	{
		double MaxValue = DBL_MIN;
		int CurrIndex_i = 0;
		int CurrIndex_j = 0;
		for(int i = 0; i < M; ++i)
		{//The JMerge Matrix is symmetric,coz i,j = j,i
			for( int j = i+1; j < M; ++j)
			{
				if(Merge_Backup[M*i+j]!= DBL_MIN && Merge_Backup[M*i+j]> MaxValue)
				{
					MaxValue = Merge_Backup[M*i+j];
					CurrIndex_i = i;
					CurrIndex_j = j;
				}
			}
		}
		Merge_Backup[M*CurrIndex_i+CurrIndex_j] = DBL_MIN;
		MaxIndex_Merge(m,0) = CurrIndex_i;
		MaxIndex_Merge(m,1) = CurrIndex_j;	
	}	

	//find k for spliting, k is chosen excluding i,j
	Matrix MaxIndex_Split(M,1);	
	for(int m = 0; m < Cmax; ++m)
	{
		int index = m%(M-2);	
		int t = round(m/(M-2));
		if(index == 0)
		{
			Split_Backup = JSplit;
		
			for(int i = 0; i < (M-2); ++i)
			{
				int CurrIndex = 0;
				double MaxValue = DBL_MIN;
				for(int j = 0; j < M; ++j)
				{
					if(j!= MaxIndex_Merge(t,0) && j!= MaxIndex_Merge(t,1)
						&& Split_Backup[j]!= DBL_MIN && Split_Backup[j]> MaxValue)
					{
						MaxValue = Split_Backup[j];
						CurrIndex = j;				
					}
				}
				Split_Backup[CurrIndex] = DBL_MIN;
				MaxIndex_Split[i] = CurrIndex;
			}
			
		}
		//find the candidates(i,j,k);
		Candidates(m,0) = MaxIndex_Merge(t,0);
		Candidates(m,1) = MaxIndex_Merge(t,1);	
		Candidates(m,2) = MaxIndex_Split[index];
	}

}


// Initilize the parameters of i,j,k 
void InitialParameters(const MixtureModel& MM, MixtureModel& NewMM, 
									int merge1, int merge2, int split)
{

	bool Diagonal = true;
    for (int k = 0; Diagonal && k < MM.Count(); ++k)
        if (MM[k].MatrixType() != MM_COV_TYPE_DIAG) Diagonal = false;
	//merge merge1 & merge2 --> index merge1
	//update weight
	double alpha1 = MM[merge1].Weight();
	double alpha2 = MM[merge2].Weight();
	double alpha3 = MM[split].Weight();
	NewMM[merge1].Weight(alpha1+alpha2);

	// update mean, covariance
	Matrix mean1 = MM[merge1].Mean();
	Matrix mean2 = MM[merge2].Mean();
	NewMM[merge1].Mean() = (alpha1/(alpha1+alpha2)) * mean1 + (alpha2/(alpha1+alpha2)) * mean2;

	Matrix cov1 = MM[merge1].Covariance();
	Matrix cov2 = MM[merge2].Covariance();
	NewMM[merge1].Covariance() = (alpha1/(alpha1+alpha2)) * cov1 + (alpha2/(alpha1+alpha2)) * cov2;
	

	//split split --> merge2 and split
	//update weight
	NewMM[merge2].Weight(alpha3/2);
	NewMM[split].Weight(alpha3/2);

	//update mean, covariance
	int d = MM.Dimensionality();
	Matrix cov3 = MM[split].Covariance();
	Matrix RandMean1(1,d);
	RandMean1.Randomize();
	Matrix RandMean2(1,d);
	RandMean2.Randomize();
	NewMM[merge2].Mean() = MM[split].Mean()+ RandMean1;
	NewMM[split].Mean() = MM[split].Mean()+ RandMean2;
	
	if(Diagonal){
		Matrix RandCov1(1,d);
		Matrix RandCov2(1,d);
		RandCov1.Randomize();
		RandCov2.Randomize();
		NewMM[merge2].Covariance() = MM[split].Covariance() + RandCov1;
		NewMM[split].Covariance() = MM[split].Covariance() + RandCov2;
	}
	else{
		Matrix RandCov1(d,d);
		RandCov1.Randomize();
		Matrix RandCov2(d,d);
		RandCov2.Randomize();
		NewMM[merge2].Covariance() = MM[split].Covariance()+ RandCov1;
		NewMM[split].Covariance() = MM[split].Covariance() + RandCov2;
	}

	/* // here is another initial way mentioned in the paper, but the result seems unreasonable
	Matrix IdentityMatrix(d,d);
	IdentityMatrix.Identity();
	IdentityMatrix *= pow(cov3.Determinant(),d);
	NewMM[merge2].Covariance() = IdentityMatrix;
	//	pow(Determinant(MM[split].Covariance()), 1/MM.Dimensionality())*Identity();
	NewMM[split].Covariance()= IdentityMatrix;
	*/
	/*
		the covariance with split initialize as c1=c2=det(c_old).^d*Id
	*/

}

// update the parameters because of the merge and split. 
// mainly formula for posteria probability
void PartialEM(const MixtureModel& MM, const Data& D, MixtureModel& NewMM,
			   int merge1, int merge2, int split, bool DiagonalOnly)
{
	Matrix xTx(DiagonalOnly ? 1 : D.Dimensionality(), D.Dimensionality());
	
	Matrix PosteriorRow(1, MM.Count());
	Matrix PosteriorRowNew(1, MM.Count());

	Matrix ColumnSums(1, MM.Count());
	ColumnSums.Fill(0.0);

	InitialParameters(MM, NewMM, merge1, merge2, split);

	SetupDeterminantAndInverseMatrices(DetInvs, MM);
	SetupDeterminantAndInverseMatrices(DetInvsNew, NewMM);

	for(size_t n= 0; n < D.Count(); ++n)
	{
		ConstRowIterator Vect = D.ConstVectorBegin(n);
		ConstRowIterator VectorEnd = D.ConstVectorEnd(n);
		double partial_sumup_old = 0;
		double partial_sumup_new = 0;
		double Denorm = 0;
		for(int k = 0; k < MM.Count(); ++k)
		{
			double Mah, Mah_new;
			if(k == merge1 || k == merge2 || k == split)
			{			
				if (MM[k].MatrixType() == MM_COV_TYPE_DIAG)
				{
					Mah = MahalanobisDiag2(Vect, MM[k].Mean().ConstRowBegin(0),
						DetInvs[k].second);
					Mah_new = MahalanobisDiag2(Vect, NewMM[k].Mean().ConstRowBegin(0),
						DetInvsNew[k].second);
				}		
				else 
				{
					Mah = MahalanobisChol2(Vect, MM[k].Mean().ConstRowBegin(0),
						DetInvs[k].second);
					Mah_new = MahalanobisChol2(Vect, NewMM[k].Mean().ConstRowBegin(0),
						DetInvsNew[k].second);
				}
				//old posterior prob.
				PosteriorRow[k] = MM[k].Weight() * ProbabilityDensity(Mah, DetInvs[k].first);
				
				partial_sumup_old += PosteriorRow[k];	
				PosteriorRowNew[k] = NewMM[k].Weight() * ProbabilityDensity(Mah_new, DetInvsNew[k].first);
				
				partial_sumup_new += PosteriorRowNew[k];
				Denorm += PosteriorRow[k];				
			}
			else
			{
				if (MM[k].MatrixType() == MM_COV_TYPE_DIAG)
					Mah = MahalanobisDiag2(Vect, MM[k].Mean().ConstRowBegin(0),
						DetInvs[k].second);
				else 
					Mah = MahalanobisChol2(Vect, MM[k].Mean().ConstRowBegin(0),
						DetInvs[k].second);
				PosteriorRow[k] = MM[k].Weight() * ProbabilityDensity(Mah, DetInvs[k].first);
				PosteriorRowNew[k] = PosteriorRow[k];
				Denorm += PosteriorRow[k];
			}
		}// end MM.Count()
		
		// factor = sum of posterior prob. for old model i,j,k 
		if (Denorm < DBL_MIN)
			partial_sumup_old = 3 / MM.Count();
		else
			partial_sumup_old = partial_sumup_old/Denorm;
				
		for(int m = 0; m< MM.Count(); ++m)
		{
			if(m == merge1 || m == merge2 || m == split)
			{
				double coef = 0;
				if(partial_sumup_new < DBL_MIN)
					coef = partial_sumup_old/ DBL_MIN;
				else
					coef = partial_sumup_old/ partial_sumup_new;
				if(coef > DBL_MAX)	
					PosteriorRowNew[m] = 1.0 / MM.Count();						
				else 
					PosteriorRowNew[m] *= coef;		
			}		
			else
			{
				if(Denorm < DBL_MIN)
					PosteriorRowNew.Fill(1.0 / MM.Count());
				else
					PosteriorRowNew[m] *= 1/Denorm;			
			}
		}

		ColumnSums += PosteriorRowNew;

		//got new posterior prob. of i,j,k; need to update weight, mean, covariance
		//no need to update weight, it's same
		//mean and covariance only for i,j,k
        // Compute outer product needed for covariance update.
    if (DiagonalOnly) {
        ConstRowIterator Curr = Vect;
        RowIterator Tgt = xTx.RowBegin(0);
        while (Curr != VectorEnd) {
            *Tgt = *Curr * *Curr;
            ++Curr;
            ++Tgt;
        }
    } else
        xTx.OuterProduct(Vect, VectorEnd, Vect, VectorEnd);
		
    for (int k = 0; k < NewMM.Count(); ++k) 
		{
			if(k == merge1 || k == merge2 || k == split)
			{
				double Prob = PosteriorRowNew[k];
				// Update mean.
				ConstRowIterator VectorCurr = Vect;
				RowIterator MeanCurr = NewMM[k].Mean().RowBegin(0);
				while (VectorCurr != VectorEnd) {
					*MeanCurr += Prob * *VectorCurr;
					++MeanCurr;
					++VectorCurr;
				}
				// Update covariance.
				if (NewMM[k].MatrixType() == MM_COV_TYPE_DIAG) 
				{
					if (DiagonalOnly)
						for (int p = 0; p < xTx.Cols(); ++p)
							NewMM[k].Covariance()[p] += Prob * xTx[p];
					else 
						for (int p = 0; p < xTx.Cols(); ++p)
							NewMM[k].Covariance()[p] += Prob * xTx(p, p);
				} 
				else for (int p = 0; p < xTx.Cols() * xTx.Rows(); ++p)
					NewMM[k].Covariance()[p] += Prob * xTx[p];
			}
    }		

	}// end D.Count()

	// Finish the update of new model.
  xTx.Resize(D.Dimensionality(), D.Dimensionality());
  for (int k = 0; k < NewMM.Count(); ++k) 
	{
		if(k == merge1 || k == merge2 || k == split)
		{
			if (ColumnSums[k] < DBL_MIN) 
				ColumnSums[k] = DBL_MIN;
			// Update weight.
			NewMM[k].Weight(ColumnSums[k] / D.Count());
			// Scale mean and covariance.
			NewMM[k].Mean() *= 1.0 / ColumnSums[k];
			NewMM[k].Covariance() *= 1.0 / ColumnSums[k];
			// Finish update of covariance by subtracting outer product of mean.
			xTx.OuterProduct(
				NewMM[k].Mean().ConstRowBegin(0), NewMM[k].Mean().ConstRowEnd(0),
				NewMM[k].Mean().ConstRowBegin(0), NewMM[k].Mean().ConstRowEnd(0));
			if (NewMM[k].MatrixType() == MM_COV_TYPE_DIAG) {
				for (size_t d = 0; d < D.Dimensionality(); ++d)
					NewMM[k].Covariance()[d] -= xTx(d, d);
			} else NewMM[k].Covariance() -= xTx;
		}
   }
	// got the updated mixture model: NewMM


}

void test(Matrix mean, Matrix cov)
{
	Matrix Mean1 = mean;
	Matrix RandMean1(1,2);
	RandMean1.Randomize();
	fprintf(stdout, "m(0,0) = %f; m(0,1) = %f\n", Mean1(0,0), Mean1(0,1));
	fprintf(stdout, "rm(0,0) = %f; rm(0,1) = %f\n", RandMean1(0,0), RandMean1(0,1));	
	Mean1 = Mean1 + RandMean1;
	fprintf(stdout, "Mean1(0,0) = %f; Mean1(0,1) = %f\n", Mean1(0,0), Mean1(0,1));
	Matrix cov1 = cov;
	fprintf(stdout, "cov(0,0) = %f; cov(0,1) = %f; cov(1,0) = %f; cov(1,1) = %f\n", cov(0,0), cov(0,1),
	            cov(1,0), cov(1,1));
	Matrix IdentityMatrix(2,2);
	IdentityMatrix.Identity();
	fprintf(stdout, "Id(0,0) = %f; Id(0,1) = %f; Id(1,0) = %f; Id(1,1) = %f\n", IdentityMatrix(0,0), IdentityMatrix(0,1),
	            IdentityMatrix(1,0), IdentityMatrix(1,1));
	IdentityMatrix *= pow(cov1.Determinant(),2);

	cov1 = IdentityMatrix;
	fprintf(stdout, "cov1(0,0) = %f; cov1(0,1) = %f; cov1(1,0) = %f; cov1(1,1) = %f\n", cov1(0,0), cov1(0,1),
	            cov1(1,0), cov1(1,1));
	
}









