/*
 * SM-EM algorithm. Either starts with random solution 
 * and returns the best.
 Qinpei Zhao 2008
 */

#include "options.h"
#include "data.hpp"
#include "mixmodel.hpp"
#include "smem.hpp"
#include "em.hpp"
#include "datautils.hpp"
#include "kmeans.hpp"
#include "timer.hpp"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#if !defined(__linux__)
#include <ieeefp.h>
#endif


static bool ParseCommandLine(int argc, char** argv) 
{
    optAdd('i', "input");
    optAddValueString(0);
    optUsage("Input data file name. One vector per row.");
    optUsage("'%' or '#' starts a comment line. Whole line is comment.");
    optCompulsory();

    optAdd('s', "solution");
    optAddValueString(0);
    optUsage("Optional initial solution.");

    optAdd('o', "output");
    optAddValueString(0);
    optCompulsory();
    optUsage("Output file name.");

    optAdd('c', "count");
    optAddValueInteger(3, 3, 1000000);
    optUsage("Number of multivariate normal distributions.");

    optAdd('v', "variance");
    optAddValueDouble(0.0000152587890625, 0.0000152587890625, 1);
    optUsage("Minimum allowed variance relative to the variance of data.");
    optUsage("Default value is 1/65536.");

    optAddFlag('f', "full", 0);
    optUsage(
        "Toggle the use of full covariance matrices, default is diagonal.");

    optAdd('r', "random");
    optAddValueInteger(10, 0, 100);
    optUsage(
        "Number of k-means iterations before SMEM for random solution.");
    optAddValueBool(1);
    optUsage("Update covariance matrices at the last k-means iteration.");

    optAdd('I', "iterations");
    optAddValueInteger(100, 0, 1000);
    optUsage("Number of SMEM iterations.");

	optAdd('E', "iterations");
    optAddValueInteger(10, 0, 1000);
    optUsage("Number of EM iterations in SMEM algorithm.");

    optAdd('n', "no.of candidates");
    optAddValueInteger(20, 0, 10000);
    optUsage("Number of candidates.");

	optAdd('t', "threshold");
    optAddValueDouble(0.0000152587890625, 0.0000152587890625, 1);
    optUsage("Ending threshold for relative improvement, default is 2^-16.");

	optAddFlag('p', "partition", 0);
    optUsage("Output partition of data instead of Gaussians.");

    optAddFlag('\0', "verbose", 0);
    optUsage("Increases the level of output, can be given several times.");
    optUsage("Giving this twice prints out variances, variance limits and");
    optUsage("not known yet.");

    optAddFlag('\0', "help", 0);
    optUsage("Print this help to stderr and exit.");

    if (optGivenOption('\0', "help", argc, argv)
        || optParseCommandLine(argc, argv, stderr)
        || (optGiven('f', 0) || optGiven('c', 0)) && optGiven('s', 0)
		|| (optGiven('E', 0) && optGiven('t', 0)))
    {
        fprintf(stderr, "\nSMEM-algorithm, version 1.0.1 (compiled %s, %s)\n",
            __DATE__, __TIME__);
        optPrintUsage(stderr, "smemax");
        fprintf(stderr,
            "You can't use --solution with --full or --count.\n");
        fprintf(stderr, "Warning about numerical problems means that the\n");
        fprintf(stderr, "log likelihood is not monotonic or not a number.\n\n");
        return false;
    }
    return true;
}

static void CheckAndWarn(const Data& D, int Components) 
{
    float Ratio = D.Count() / (float)Components;
    if (Ratio <= D.Dimensionality()) 
	{
        fprintf(stderr, "\nWarning: probably too few data vectors (%i)",
            D.Count());
        fprintf(stderr, " or too many components (%i)\n", Components);
        fprintf(stderr, "  for data of dimensionality %i.\n",
            D.Dimensionality());
        fprintf(stderr, "  Data / components ratio is %f.\n\n", Ratio);
    } 
	else if (Ratio < 1.5f * D.Dimensionality()) 
	{
        fprintf(stderr, "\nWarning: maybe too few data vectors (%i)",
            D.Count());
        fprintf(stderr, " or too many components (%i)\n", Components);
        fprintf(stderr, "  for data of dimensionality %i.\n",
            D.Dimensionality());
        fprintf(stderr, "  Data / components ratio is %f.\n\n", Ratio);
    }
}

int main(int argc, char** argv) 
{
    if (!ParseCommandLine(argc, argv)) return 1;
    srand((getpid() << 15) + getppid());
    DataStorage DS;
    if (DS.Read(optString('i', 1)))
	{
        fprintf(stderr, "\nError reading input data file: %s\n\n",
            optString('i', 1));
        return 2;
    }
    Data D(&DS);
    D.Map(0, DS.Count());
    Matrix Cov;
    bool VeryVerbose = optGivenStr("verbose", 0) > 1;
    if (VeryVerbose) 
	{
        Matrix Mean = DataMean(D);
        fprintf(stdout, "Data mean:\n");
        Mean.Print(stdout);
        Cov = DataCovariance(D, Mean);
    } else Cov = DataCovariance(D, DataMean(D));
    Matrix MinCov;
    Cov.Diagonal(MinCov);
    if (optGivenStr("verbose", 0) > 1) 
	{
        fprintf(stdout, "Data variances:\n");
        MinCov.Print(stdout);
    }
    if (!optBool('f', 0)) Cov = MinCov;
    Cov *= 1.0 / optInt('c', 1);
    MinCov *= optDouble('v', 1);
    if (optGivenStr("verbose", 0) > 1) 
	{
        fprintf(stdout, "Minimum variances:\n");
        MinCov.Print(stdout);
    }


	//fprintf(stdout, "cmax:%d\n", optInt('n', 1));

    MixtureModel MM;
    double LogLikelihood = 0;
	int m = 0; // model size
	  //add by qinpei
	  double Qvalue = 0;
    if (optGiven('s', 0)) 
	{		
        MixtureModelFileFormat MMFF;
        if (MM.Read(optString('s', 1), MMFF)) 
				{
            fprintf(stderr, "\nError reading input solution: %s\n",
                optString('s', 1));
            return 3;
        }
        CheckAndWarn(D, MM.Count());
		m = MM.Count();
        MM.LimitVariances(MinCov);
        MM.NormalizeWeights();
		if(optGiven('t',0))
			Qvalue = SplitMergeEM(MM, D, MinCov, optInt('I', 1), optDouble('t', 1), optInt('n', 1), VeryVerbose);
		else
			Qvalue = SplitMergeEM(MM, D, MinCov, optInt('I', 1), optInt('E', 1), optInt('n', 1), VeryVerbose);

    } 	
	else {
        CheckAndWarn(D, optInt('c', 1));
        for (int k = 0; k < optInt('c', 1); ++k)
            MM.Add(*(new MixtureComponent(
                optBool('f', 0) ? MM_COV_TYPE_FULL : MM_COV_TYPE_DIAG,
                D.Dimensionality())));
        std::vector<int> Indices;
        if (!SelectUniquePoints(D, MM.Count(), Indices)) 
				{
            fprintf(stderr,
                "\nError: more components than unique points (%i).\n",
                Indices.size());
            return 5;
        }
        for (int n = 0; n < MM.Count(); ++n) 
		{
            MM[n].Weight(1.0 / MM.Count());
            MM[n].Mean(D.ConstVectorBegin(Indices[n]));
            MM[n].Covariance() = Cov;
		}
//		MM.Write(optString('o', 1), MM_FF_PLAIN);
//		fprintf(stdout, "finish writing initialized models\n");
        KMeans(D, MM, optInt('r', 1), optBool('r', 2) != 0);
//        print out the MM here!!!
		if(optGiven('t',0))
			Qvalue = SplitMergeEM(MM, D, MinCov, optInt('I', 1), 
					optDouble('t', 1), optInt('n', 1), VeryVerbose);
		else
			Qvalue = SplitMergeEM(MM, D, MinCov, optInt('I', 1), 
					optInt('E', 1), optInt('n', 1),  VeryVerbose);

		m = optInt('c', 1);
    }
    if (optGivenStr("verbose", 0)) 
		{
        fprintf(stdout, "CPU time: %s\n", timer::StringS(timer::Now()));
        fprintf(stdout, "Final result log likelihood: ");
    }
    fprintf(stdout, "%g\n\n", Qvalue);
	
	// Create and write partition
    if (optGiven('p', 0)) {
      std::vector<int> Partition(DS.Count(), -1);
      emBayesPartitioning(D, MM, Partition);
      
      FILE* f = fopen(optString('o', 1), "w");
      if (f == NULL) {
		fprintf(stderr, "\nError writing output to: %s\n",
			optString('o', 1));
		return 4;
      }

      for (size_t i = 0; i < Partition.size(); ++i) {
		fprintf(f, "%d\n", Partition[i]);
      }
      fclose(f);
	}
     else if ( optInt('I', 1)) 
	{
        if (MM.Write(optString('o', 1), MM_FF_PLAIN)) {             fprintf(stderr, "\nError writing output to: %s\n",                 optString('o', 1));             return 4; 		}     }
    return 0;
}


